"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const CBOR = require("cbor");
const knownSections = [
    'critical',
    'index',
    'manifest',
    'responses',
    'signatures',
];
/** This class represents parsed Web Bundle. */
class Bundle {
    constructor(buffer) {
        this.sections = {};
        this.responses = {}; // Offset-in-responses -> resp
        const wbn = asArray(CBOR.decode(buffer));
        if (wbn.length !== 6) {
            throw new Error('Wrong toplevel structure');
        }
        const [magic, version, primaryURL, sectionLengthsCBOR, sections, length,] = wbn;
        if (bytestringToString(magic) !== '🌐📦') {
            throw new Error('Wrong magic');
        }
        this.version = bytestringToString(version).replace(/\0+$/, ''); // Strip off the '\0' paddings.
        this.primaryURL = asString(primaryURL);
        const sectionLengths = asArray(CBOR.decode(asBytestring(sectionLengthsCBOR)));
        const sectionsArray = asArray(sections);
        if (sectionLengths.length !== sectionsArray.length * 2) {
            throw new Error("Number of elements in section-lengths and in sections don't match");
        }
        for (let i = 0; i < sectionsArray.length; i++) {
            this.sections[asString(sectionLengths[i * 2])] = sectionsArray[i];
        }
        if (this.sections['critical']) {
            for (const name of asArray(this.sections['critical'])) {
                if (!knownSections.includes(asString(name))) {
                    throw new Error(`unknown section ${name} is marked as critical`);
                }
            }
        }
        // The index section records (offset, length) of each response, but our
        // CBOR decoder doesn't preserve location information. So, recalculate
        // offset and length of each response here. This is inefficient, but works.
        const responses = asArray(this.sections['responses']);
        let offsetInResponses = encodedLength(responses.length);
        for (const resp of responses) {
            this.responses[offsetInResponses] = new Response(asArray(resp));
            offsetInResponses += encodedLength(resp);
        }
    }
    get manifestURL() {
        if (this.sections['manifest']) {
            return asString(this.sections['manifest']);
        }
        return null;
    }
    get urls() {
        return Object.keys(this.indexSection);
    }
    getResponse(url) {
        const indexEntry = asArray(this.indexSection[url]);
        if (!indexEntry) {
            throw new Error('No entry for ' + url);
        }
        const [variants, offset, length] = indexEntry;
        if (asBytestring(variants).length !== 0) {
            throw new Error('Variants are not supported');
        }
        if (indexEntry.length !== 3) {
            throw new Error('Unexpected length of index entry for ' + url);
        }
        const resp = this.responses[asNumber(offset)];
        if (!resp) {
            throw new Error(`Response for ${url} is not found (broken index)`);
        }
        return resp;
    }
    get indexSection() {
        return asMap(this.sections['index']);
    }
}
exports.Bundle = Bundle;
/** This class represents an HTTP resource in Web Bundle. */
class Response {
    constructor(responsesSectionItem) {
        if (responsesSectionItem.length !== 2) {
            throw new Error('Wrong response structure');
        }
        const { status, headers } = decodeResponseMap(asBytestring(responsesSectionItem[0]));
        this.status = status;
        this.headers = headers;
        this.body = asBytestring(responsesSectionItem[1]);
    }
}
exports.Response = Response;
function encodedLength(value) {
    return CBOR.encode(value).byteLength;
}
function decodeResponseMap(cbor) {
    const decoded = CBOR.decode(cbor);
    if (!(decoded instanceof Map)) {
        throw new Error('Wrong header map structure');
    }
    let status = null;
    const headers = {};
    for (let [key, val] of decoded.entries()) {
        key = bytestringToString(key);
        val = bytestringToString(val);
        if (key === ':status') {
            status = Number(val);
        }
        else if (key.startsWith(':')) {
            throw new Error('Unknown psuedo header ' + key);
        }
        else {
            headers[key] = val;
        }
    }
    if (!status) {
        throw new Error('No :status in response header map');
    }
    return { status, headers };
}
// Type assertions and conversions for CBOR-decoded objects.
function asArray(x) {
    if (x instanceof Array) {
        return x;
    }
    throw new Error('Array expected, but got ' + typeof x);
}
function asMap(x) {
    if (typeof x === 'object' && x !== null && !(x instanceof Array)) {
        return x;
    }
    throw new Error('Map expected, but got ' + typeof x);
}
function asNumber(x) {
    if (typeof x === 'number') {
        return x;
    }
    throw new Error('Number expected, but got ' + typeof x);
}
function asString(x) {
    if (typeof x === 'string') {
        return x;
    }
    throw new Error('String expected, but got ' + typeof x);
}
function asBytestring(x) {
    if (x instanceof Buffer) {
        return x;
    }
    throw new Error('Bytestring expected, but got ' + typeof x);
}
function bytestringToString(bstr) {
    if (!(bstr instanceof Buffer)) {
        throw new Error('Bytestring expected');
    }
    return bstr.toString('utf-8');
}
