"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const CBOR = require("cbor");
const fs = require("fs");
const mime = require("mime");
const path = require("path");
const url_1 = require("url");
class BundleBuilder {
    constructor(primaryURL) {
        this.primaryURL = primaryURL;
        this.sectionLengths = [];
        this.sections = [];
        this.responses = [];
        this.index = new Map();
        this.currentResponsesOffset = 0;
        validateExchangeURL(primaryURL);
    }
    // TODO: Provide async version of this.
    createBundle() {
        if (!this.index.has(this.primaryURL)) {
            throw new Error(`Exchange for primary URL (${this.primaryURL}) does not exist`);
        }
        this.addSection('index', this.fixupIndex());
        this.addSection('responses', this.responses);
        const estimatedBundleSize = this.sectionLengths.reduce((size, s) => size + s.length, 16384 // For headers (including primary URL)
        );
        const wbn = encodeCanonical(this.createTopLevel(), estimatedBundleSize);
        // Fill in the length field.
        const view = new DataView(wbn.buffer, wbn.byteOffset + wbn.length - 8);
        view.setUint32(0, Math.floor(wbn.length / 0x100000000));
        view.setUint32(4, wbn.length & 0xffffffff);
        return wbn;
    }
    addExchange(url, status, headers, payload) {
        validateExchangeURL(url);
        if (typeof payload === 'string') {
            payload = byteString(payload);
        }
        this.addIndexEntry(url, this.addResponse(new HeaderMap(status, headers), payload));
        return this;
    }
    addFile(url, file) {
        const headers = {
            'Content-Type': mime.getType(file) || 'application/octet-stream',
        };
        this.addExchange(url, 200, headers, fs.readFileSync(file));
        return this;
    }
    addFilesRecursively(baseURL, dir) {
        if (!baseURL.endsWith('/')) {
            throw new Error("baseURL must end with '/'.");
        }
        const files = fs.readdirSync(dir);
        files.sort(); // Sort entries for reproducibility.
        for (const file of files) {
            const filePath = path.join(dir, file);
            if (fs.statSync(filePath).isDirectory()) {
                this.addFilesRecursively(baseURL + file + '/', filePath);
            }
            else if (file === 'index.html') {
                // If the file name is 'index.html', create an entry for baseURL itself
                // and another entry for baseURL/index.html which redirects to baseURL.
                // This matches the behavior of gen-bundle.
                this.addFile(baseURL, filePath);
                this.addExchange(baseURL + file, 301, { Location: './' }, '');
            }
            else {
                this.addFile(baseURL + file, filePath);
            }
        }
        return this;
    }
    setManifestURL(url) {
        validateExchangeURL(url);
        this.addSection('manifest', url);
        return this;
    }
    addSection(name, content) {
        if (this.sectionLengths.some(s => s.name === name)) {
            throw new Error('Duplicated section: ' + name);
        }
        let length;
        if (name === 'responses') {
            // The responses section can be large, so avoid using encodedLength()
            // with the entire section's content.
            // Here, this.currentResponsesOffset holds the sum of all response
            // lengths. Adding encodedLength(this.responses.length) to this gives
            // the same result as encodedLength(content).
            length =
                this.currentResponsesOffset + encodedLength(this.responses.length);
        }
        else {
            length = encodedLength(content);
        }
        this.sectionLengths.push({ name, length });
        this.sections.push(content);
    }
    // Adds a response to `this.response`, and returns its length in the
    // responses section.
    addResponse(headerMap, payload) {
        if (payload.length > 0 && !headerMap.has('content-type')) {
            throw new Error('Non-empty exchange must have Content-Type header');
        }
        const response = [new Uint8Array(encodeCanonical(headerMap)), payload];
        this.responses.push(response);
        // This should be the same as encodedLength(response).
        return response.reduce((len, buf) => len + encodedLength(buf.length) + buf.length, 1);
    }
    addIndexEntry(url, responseLength) {
        this.index.set(url, [
            new Uint8Array(0),
            this.currentResponsesOffset,
            responseLength,
        ]);
        this.currentResponsesOffset += responseLength;
    }
    fixupIndex() {
        // Adjust the offsets by the length of the response section's CBOR header.
        const responsesHeaderSize = encodedLength(this.responses.length);
        for (const value of this.index.values()) {
            value[1] += responsesHeaderSize;
        }
        return this.index;
    }
    createTopLevel() {
        const sectionLengths = [];
        for (const s of this.sectionLengths) {
            sectionLengths.push(s.name, s.length);
        }
        return [
            byteString('🌐📦'),
            byteString('b1\0\0'),
            this.primaryURL,
            new Uint8Array(encodeCanonical(sectionLengths)),
            this.sections,
            new Uint8Array(8),
        ];
    }
}
exports.BundleBuilder = BundleBuilder;
class HeaderMap extends Map {
    constructor(status, headers) {
        super();
        if (status < 100 || status > 999) {
            throw new Error('Invalid status code');
        }
        this.set(':status', status.toString());
        for (const key of Object.keys(headers)) {
            this.set(key.toLowerCase(), headers[key]);
        }
    }
    // This tells the CBOR library how to serialize this object.
    encodeCBOR(encoder) {
        // Convert keys and values to Uint8Array, as the CBOR representation of
        // header map is {bytestring => bytestring}.
        const m = new Map();
        for (const [key, value] of this.entries()) {
            m.set(byteString(key), byteString(value));
        }
        return encoder.pushAny(m);
    }
}
// Throws an error if `urlString` is not a valid exchange URL, i.e. it must:
// - be absolute, and
// - have no credentials (user:password@) or hash (#fragment).
function validateExchangeURL(urlString) {
    const url = new url_1.URL(urlString); // This throws if urlString is not an absolute URL.
    if (url.username !== '' || url.password !== '') {
        throw new Error('Exchange URL must not have credentials: ' + urlString);
    }
    if (url.hash !== '') {
        throw new Error('Exchange URL must not have a hash: ' + urlString);
    }
}
// Encodes `value` in canonical CBOR. Throws an error if the result is larger
// than `bufferSize`, because the CBOR encoder silently ignores write errors
// to its internal stream and returns broken data.
function encodeCanonical(value, bufferSize = 1024 * 1024) {
    const buf = CBOR.encodeOne(value, {
        canonical: true,
        highWaterMark: bufferSize,
    });
    if (buf.length >= bufferSize) {
        throw new Error('CBOR encode error: insufficient buffer size');
    }
    return buf;
}
// Returns the length of `value` when CBOR-encoded.
function encodedLength(value, bufferSize = 1024 * 1024) {
    // We don't need to use canonical encoding here.
    const len = CBOR.encodeOne(value, { highWaterMark: bufferSize }).byteLength;
    if (len >= bufferSize) {
        throw new Error('CBOR encode error: insufficient buffer size');
    }
    return len;
}
function byteString(s) {
    return Buffer.from(s, 'utf-8');
}
